/*
 * kgraft_patch_bsc1090368
 *
 * Fix for CVE-2018-8897, bsc#1090368
 *
 *  Upstream commit:
 *  d8ba61ba58c8 ("x86/entry/64: Don't use IST entry for #BP stack")
 *
 *  SLE12 commit:
 *  d82457ccb4d94672995489eb1cb0b087510cbf9e
 *
 *  SLE12-SP1 commit
 *  ed7901eb499f3092bad6285f51228627ba901e2b
 *
 *  SLE12-SP2 commit:
 *  409b5f7eb20038c99154633dccb691d8ffc2bf63
 *
 *  SLE12-SP3 commit:
 *  95d3e6c3bd7ee22b1bf33b4e4f91eedd3dc26f0a ("stable 4.4.125")
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_X86_64) && !IS_ENABLED(CONFIG_XEN)

#define pr_fmt(fmt) "kgraft-patch: " fmt

#include <linux/kernel.h>
#include <linux/kallsyms.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/atomic.h>
#include <linux/workqueue.h>
#include <linux/irqflags.h>
#include <linux/delay.h>
#include <asm/desc.h>
#include <asm/irq_vectors.h>
#include <asm/traps.h>
#include <asm/pgtable_types.h>
#include <asm/cmpxchg.h>
#include <asm/fixmap.h>

#include "patch_state.h"
#include "kgr_patch_entry_code.h"
#include "kgr_patch_bsc1090368.h"

#define BSC1090368_SUBPATCH_ID KGR_SUBPATCH_ID_BSC(1090368)

#if !IS_ENABLED(CONFIG_TRACING)
#error "Livepatch only supports CONFIG_TRACING=y"
#endif


static gate_desc (*kgr_orig_idt_table)[NR_VECTORS];
static gate_desc (*kgr_orig_trace_idt_table)[NR_VECTORS];

static struct desc_ptr *kgr_idt_descr;
static struct desc_ptr *kgr_trace_idt_descr;
static const struct desc_ptr *kgr_debug_idt_descr;

static u32 __percpu *kgr_debug_idt_ctr;
static atomic_t *kgr_trace_idt_ctr;

char (*kgr_error_entry)[];
char (*kgr_error_exit)[];
static asmlinkage void (*kgr_orig_int3)(void);
dotraplinkage void (*kgr_do_int3)(struct pt_regs *regs, long error_code);

int (*kgr_schedule_on_each_cpu)(work_func_t func);

#if IS_ENABLED(CONFIG_KAISER)
int (*kgr_kaiser_add_mapping)(unsigned long addr, unsigned long size,
			      unsigned long flags);
void (*kgr_kaiser_remove_mapping)(unsigned long start, unsigned long size);
void (*kgr_flush_tlb_kernel_range)(unsigned long start, unsigned long end);
#endif


static struct {
	char *name;
	char **addr;
} kgr_funcs[] = {
	{ "idt_table", (void *)&kgr_orig_idt_table },
	{ "trace_idt_table", (void *)&kgr_orig_trace_idt_table },
	{ "idt_descr", (void *)&kgr_idt_descr },
	{ "trace_idt_descr", (void *)&kgr_trace_idt_descr },
	{ "debug_idt_descr", (void *)&kgr_debug_idt_descr },
	{ "debug_idt_ctr", (void *)&kgr_debug_idt_ctr },
	{ "trace_idt_ctr", (void *)&kgr_trace_idt_ctr },
	{ "error_entry", (void *)&kgr_error_entry },
	{ "error_exit", (void *)&kgr_error_exit },
	{ "int3", (void *)&kgr_orig_int3 },
	{ "do_int3", (void *)&kgr_do_int3 },
	{ "schedule_on_each_cpu", (void *)&kgr_schedule_on_each_cpu },
#if IS_ENABLED(CONFIG_KAISER)
	{ "kaiser_add_mapping", (void *)&kgr_kaiser_add_mapping },
	{ "kaiser_remove_mapping", (void *)&kgr_kaiser_remove_mapping },
	{ "flush_tlb_kernel_range", (void *)&kgr_flush_tlb_kernel_range },
#endif
};



/* from arch/x86/include/asm/desc.h */
static inline bool kgr_is_debug_idt_enabled(void)
{
	if (*this_cpu_ptr(kgr_debug_idt_ctr))
		return true;

	return false;
}

static inline void kgr_load_debug_idt(void)
{
	load_idt((const struct desc_ptr *)kgr_debug_idt_descr);
}

static inline bool kgr_is_trace_idt_enabled(void)
{
	if (atomic_read(kgr_trace_idt_ctr))
		return true;

	return false;
}

static inline void kgr_load_trace_idt(void)
{
	load_idt((const struct desc_ptr *)kgr_trace_idt_descr);
}

static inline void kgr_load_current_idt(void)
{
	if (kgr_is_debug_idt_enabled())
		kgr_load_debug_idt();
	else if (kgr_is_trace_idt_enabled())
		kgr_load_trace_idt();
	else
		load_idt((const struct desc_ptr *)kgr_idt_descr);
}



static void __kgr_sync_idtr(struct work_struct *work)
{
	unsigned long flags;

	local_irq_save(flags);
	kgr_load_current_idt();
	/*
	 * At this point, we're in process context
	 * and thus, any interrupts or exceptions entered
	 * have finished executing.
	 */
	local_irq_restore(flags);
}

static void kgr_sync_idtr(bool lock_mod_on_fail)
{
	int retries;
	int ret;

	ret = kgr_schedule_on_each_cpu(__kgr_sync_idtr);
	if (ret) {
		/*
		 * Uh oh, memory allocation failure, keep retrying for
		 * some time.
		 */
		pr_warn("sync_idtr failed, keep retrying for 120s...\n");
		for (retries = 0; ret && retries < 120; ++retries) {
			msleep(1000);
			ret = kgr_schedule_on_each_cpu(__kgr_sync_idtr);
		}

		if (!ret) {
			pr_info("sync_idtr succeeded\n");
		} else {
			if (lock_mod_on_fail) {
				pr_err("sync_idtr failed, module removal prevented\n");
				/* Leak a reference on self. */
				__module_get(THIS_MODULE);
			} else {
				pr_warn("sync_idtr failed\n");
			}
		}
	}
}

static gate_desc (*kgr_repl_idt_table)[NR_VECTORS];
static gate_desc (*kgr_repl_trace_idt_table)[NR_VECTORS];

extern asmlinkage void kgr_repl_int3(void);

static bool replaced_idt = false;

static void bsc1090368_post_patch(struct kgr_subpatch_data *header)
{
	gate_desc expected_bp_gate;
	gate_desc repl_bp_gate;

	/*
	 * Safety check: verify that the current IDT entry is what
	 * it's expected to be.
	 *
	 * The X86_TRAP_BP IDT entry is expected to have been setup by
	 *  trap_init()
	 *   ->set_system_intr_gate_ist(X86_TRAP_BP, &int3, DEBUG_STACK)
	 *    ->_set_gate(n, GATE_INTERRUPT, addr, 0x3, ist, __KERNEL_CS);
	 *     ->pack_gate(&s, type, (unsigned long)addr, dpl, ist, seg);
	 * Verify that this has indeed been the case.
	 */
	pack_gate(&expected_bp_gate, GATE_INTERRUPT,
		  (unsigned long)kgr_orig_int3, 0x3, DEBUG_STACK, __KERNEL_CS);

	if (memcmp(&(*kgr_orig_idt_table)[X86_TRAP_BP], &expected_bp_gate,
		   sizeof(expected_bp_gate)) ||
	    memcmp(&(*kgr_orig_trace_idt_table)[X86_TRAP_BP], &expected_bp_gate,
		   sizeof(expected_bp_gate))) {
		pr_warn("BP IDT entry has unexpected value, NOT patching");
		return;
	}

	/* Prepare the IDT table replacements */
	memcpy(kgr_repl_idt_table, kgr_orig_idt_table,
	       sizeof(*kgr_repl_idt_table));
	memcpy(kgr_repl_trace_idt_table, kgr_orig_idt_table,
	       sizeof(*kgr_repl_trace_idt_table));

	pack_gate(&repl_bp_gate, GATE_INTERRUPT,
		  (unsigned long)kgr_repl_int3, 0x3, 0, __KERNEL_CS);
	write_idt_entry(*kgr_repl_idt_table, X86_TRAP_BP, &repl_bp_gate);
	write_idt_entry(*kgr_repl_trace_idt_table, X86_TRAP_BP, &repl_bp_gate);

	smp_wmb();

	/*
	 * Install the IDT table replacements.
	 * Note that struct desc_ptr's ->address is not aligned,
	 * so read and write it with a lock prefix.
	 */
	if (cmpxchg(&kgr_idt_descr->address,
		    __fix_to_virt(FIX_RO_IDT),
		    (unsigned long)*kgr_repl_idt_table) !=
	    __fix_to_virt(FIX_RO_IDT)) {
		pr_warn("idt_descr has unexpected address, NOT patching");
		return;
	}

	if (cmpxchg(&kgr_trace_idt_descr->address,
		    (unsigned long)*kgr_orig_trace_idt_table,
		    (unsigned long)*kgr_repl_trace_idt_table) !=
	    (unsigned long)*kgr_orig_trace_idt_table) {
		pr_warn("trace_idt_descr has unexpected address, NOT patching");
		xchg(&kgr_idt_descr->address, __fix_to_virt(FIX_RO_IDT));
		return;
	}

	replaced_idt = true;

	kgr_sync_idtr(false);
}

static void bsc1090368_pre_revert(struct kgr_subpatch_data *header)
{
	if (!replaced_idt)
		return;

	/* xchg() for atomic write to unaligned ->address */
	xchg(&kgr_idt_descr->address, __fix_to_virt(FIX_RO_IDT));
	xchg(&kgr_trace_idt_descr->address,
	     (unsigned long)*kgr_orig_trace_idt_table);

	kgr_sync_idtr(true);
}

static struct kgr_subpatch bsc1090368_subpatch = {
	.mod = THIS_MODULE,

	/*
	 * There is no shared data or something like a clean handover:
	 * the entry text will go away on rmmod.
	 */
	.alloc_data = NULL,
	.free_data = NULL,
	.prepare_migration = NULL, /* Migration will never happen. */

	.post_patch = bsc1090368_post_patch,
	.pre_revert = bsc1090368_pre_revert,
};

static int kgr_patch_bsc1090368_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int __init kgr_patch_bsc1090368_init(void)
{
	int ret;

	ret =  kgr_patch_bsc1090368_kallsyms();
	if (ret)
		return ret;

	ret = kgr_patch_entry_code_init();
	if (ret)
		return ret;

	/* KAISER-map the entry text. */
#if IS_ENABLED(CONFIG_KAISER)
	ret = kgr_kaiser_add_mapping(kgr_entry_text_begin,
				     kgr_entry_text_end - kgr_entry_text_begin,
				     __PAGE_KERNEL_RX);
	if (ret)
		return ret;
	kgr_flush_tlb_kernel_range(kgr_entry_text_begin, kgr_entry_text_end);
#endif

	/* Allocate and KAISER-map the IDT table replacements. */
	kgr_repl_idt_table = vzalloc(sizeof(*kgr_repl_idt_table));
	if (!kgr_repl_idt_table) {
		ret = -ENOMEM;
		goto err_kaiser_unmap_entry_text;
	}

#if IS_ENABLED(CONFIG_KAISER)
	ret = kgr_kaiser_add_mapping((unsigned long)kgr_repl_idt_table,
				     sizeof(*kgr_repl_idt_table),
				     __PAGE_KERNEL);
	if (ret)
		goto err_free_repl_idt_table;
	kgr_flush_tlb_kernel_range((unsigned long)kgr_repl_idt_table,
				   ((unsigned long)kgr_repl_idt_table +
				    sizeof(*kgr_repl_idt_table)));
#endif

	kgr_repl_trace_idt_table = vzalloc(sizeof(*kgr_repl_trace_idt_table));
	if (!kgr_repl_trace_idt_table) {
		ret = -ENOMEM;
		goto err_kaiser_unmap_repl_idt_table;
	}

#if IS_ENABLED(CONFIG_KAISER)
	ret = kgr_kaiser_add_mapping((unsigned long)kgr_repl_trace_idt_table,
				     sizeof(*kgr_repl_trace_idt_table),
				     __PAGE_KERNEL);
	if (ret)
		goto err_free_repl_trace_idt_table;
	kgr_flush_tlb_kernel_range((unsigned long)kgr_repl_trace_idt_table,
				   ((unsigned long)kgr_repl_trace_idt_table +
				    sizeof(*kgr_repl_trace_idt_table)));
#endif

	ret = kgr_subpatch_register(BSC1090368_SUBPATCH_ID,
				    &bsc1090368_subpatch);
	if (ret)
		goto err_kaiser_unmap_repl_trace_idt_table;


	return 0;

	/*
	 * Error handling is a PITA because kgr_kaiser_remove_mapping()
	 * must not be called on non-mapped ranges.
	 */
err_kaiser_unmap_repl_trace_idt_table:
#if IS_ENABLED(CONFIG_KAISER)
	kgr_kaiser_remove_mapping((unsigned long)kgr_repl_trace_idt_table,
				  sizeof(*kgr_repl_trace_idt_table));
	kgr_flush_tlb_kernel_range((unsigned long)kgr_repl_trace_idt_table,
				   ((unsigned long)kgr_repl_trace_idt_table +
				    sizeof(*kgr_repl_trace_idt_table)));
#endif

err_free_repl_trace_idt_table:
	vfree(kgr_repl_trace_idt_table);

err_kaiser_unmap_repl_idt_table:
#if IS_ENABLED(CONFIG_KAISER)
	kgr_kaiser_remove_mapping((unsigned long)kgr_repl_idt_table,
				  sizeof(*kgr_repl_idt_table));
	kgr_flush_tlb_kernel_range((unsigned long)kgr_repl_idt_table,
				   ((unsigned long)kgr_repl_idt_table +
				    sizeof(*kgr_repl_idt_table)));
#endif

err_free_repl_idt_table:
	vfree(kgr_repl_idt_table);

err_kaiser_unmap_entry_text:
#if IS_ENABLED(CONFIG_KAISER)
	kgr_kaiser_remove_mapping(kgr_entry_text_begin,
				  kgr_entry_text_end - kgr_entry_text_begin);
	kgr_flush_tlb_kernel_range(kgr_entry_text_begin, kgr_entry_text_end);
#endif
	return ret;
}

void kgr_patch_bsc1090368_cleanup(void)
{
	kgr_subpatch_unregister(&bsc1090368_subpatch);

#if IS_ENABLED(CONFIG_KAISER)
	kgr_kaiser_remove_mapping((unsigned long)kgr_repl_trace_idt_table,
				  sizeof(*kgr_repl_trace_idt_table));
	kgr_flush_tlb_kernel_range((unsigned long)kgr_repl_trace_idt_table,
				   ((unsigned long)kgr_repl_trace_idt_table +
				    sizeof(*kgr_repl_trace_idt_table)));
#endif

	vfree(kgr_repl_trace_idt_table);

#if IS_ENABLED(CONFIG_KAISER)
	kgr_kaiser_remove_mapping((unsigned long)kgr_repl_idt_table,
				  sizeof(*kgr_repl_idt_table));
	kgr_flush_tlb_kernel_range((unsigned long)kgr_repl_idt_table,
				   ((unsigned long)kgr_repl_idt_table +
				    sizeof(*kgr_repl_idt_table)));
#endif

	vfree(kgr_repl_idt_table);

#if IS_ENABLED(CONFIG_KAISER)
	kgr_kaiser_remove_mapping(kgr_entry_text_begin,
				  kgr_entry_text_end - kgr_entry_text_begin);
	kgr_flush_tlb_kernel_range(kgr_entry_text_begin, kgr_entry_text_end);
#endif
}

#endif /* IS_ENABLED(CONFIG_X86_64) && !IS_ENABLED(CONFIG_XEN) */
