/*
 * kgraft_patch_entry_code.c
 *
 * Apply custom relocations to referencing kallsyms-resolved addresses
 * to entry code replacement.
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_X86_64) && !IS_ENABLED(CONFIG_XEN)

#define pr_fmt(fmt) "kgraft-patch: " fmt

#include <linux/kernel.h>
#include <linux/kallsyms.h>
#include <linux/init.h>
#include <linux/printk.h>
#include <linux/module.h>
#include <asm/cacheflush.h>
#include "kgr_patch_entry_code.h"


static int (*kgr_core_kernel_text)(unsigned long addr) __initdata ;

static struct {
	char *name;
	char **addr;
} kgr_funcs[] __initdata = {
	{ "core_kernel_text", (void *)&kgr_core_kernel_text },
};

/**
 * struct kgr_call_reloc - relocation record for entry code replacement
 * @location: Address of displacement in call or jmp isns to patch.
 * @address: Pointer to storage holding the address to call or jmp
 *           to. The additional indirection comes handy with
 *           kallsyms-resolved symbols.
 *
 * &struct kgr_call_reloc records are created from assembly,
 * c.f. KGR_CALL_PATCH and KGR_JMP_PATCH.
 */
struct kgr_call_reloc
{
	unsigned long location;
	unsigned long *address;
};

extern const struct kgr_call_reloc kgr_call_relocs_begin[] __initconst;
extern const struct kgr_call_reloc kgr_call_relocs_end[] __initconst;


/* stolen from kernel/module.c */
static inline int
kgr_within(unsigned long addr, void *start, unsigned long size)
{
	return ((void *)addr >= start && (void *)addr < start + size);
}

/*
 * Check that a given address is within this module's .text,
 * c.f. is_module_text_address().
 */
static int __init kgr_this_module_text(unsigned long addr)
{
	return kgr_within(addr, THIS_MODULE->module_core,
			  THIS_MODULE->core_text_size);
}

/*
 * Check that a given address range is contained within
 * the entry code replacement.
 */
static int __init kgr_this_module_entry_text(unsigned long addr,
					     unsigned int size)
{
  return (kgr_entry_text_begin <= addr && addr <= kgr_entry_text_end - size);
}

static unsigned int __init
kgr_call_reloc_displ(const struct kgr_call_reloc *rel)
{
	unsigned long address;

	/* relative to next insn's %rip */
	address = rel->location + 4;
	address = *rel->address - address;

	/*
	 * disp32 gets sign-extended. Check that all 32 high bits
	 * match the sign bit, i.e. bit 31.
	 */
	if (((address & (1UL << 31)) && (address >> 32 != ~0U)) ||
	     (!(address & (1UL << 31)) && (address >> 32 != 0U))) {
		pr_err("relative displacement exceeds 32 bits\n");
		return 0;
	}
	return (unsigned int)address;
}

/**
 * kgr_patch_entry_code_relocate() - apply entry text relocations
 *
 * Return:
 * 0 - OK
 * -EINVAL - fishy relocation record encountered
 * other - set_memory_rw() or set_memory_ro() failed.
 */
static int __init kgr_patch_entry_code_relocate(void)
{
	const int nr_entry_text_pages =
		(kgr_entry_text_end - kgr_entry_text_begin) >> PAGE_SHIFT;
	const struct kgr_call_reloc *call_reloc;
	int ret;

	/* Sanity check: make sure that all addresses are valid */
	if ((kgr_entry_text_begin & ~PAGE_MASK) ||
	    (kgr_entry_text_end & ~PAGE_MASK)) {
		pr_err("unexpected alignment of KGraft entry code\n");
		return -EINVAL;
	}

	for (call_reloc = kgr_call_relocs_begin;
	     call_reloc != kgr_call_relocs_end; ++call_reloc) {
		if (!kgr_this_module_entry_text(call_reloc->location, 4)) {
			pr_err("call patch location not in entry code\n");
			return -EINVAL;
		}

		if (!kgr_this_module_text(*call_reloc->address) &&
		    !kgr_core_kernel_text(*call_reloc->address)) {
			pr_err("attempted to patch call to non-text address\n");
			return -EINVAL;
		}

		if (!kgr_call_reloc_displ(call_reloc))
			return -EINVAL;
	}

	/* Actually apply the relocations to the entry text */
	ret = set_memory_rw(kgr_entry_text_begin, nr_entry_text_pages);
	if (ret)
		return ret;

	for (call_reloc = kgr_call_relocs_begin;
	     call_reloc != kgr_call_relocs_end; ++call_reloc) {
		const unsigned int displ = kgr_call_reloc_displ(call_reloc);
		memcpy((void*)call_reloc->location, &displ, sizeof(displ));
	}

	ret = set_memory_ro(kgr_entry_text_begin, nr_entry_text_pages);
	if (ret)
		return ret;

	/* This is a nop on x86, but include it here for documentation. */
	flush_icache_range(kgr_entry_text_begin, kgr_entry_text_end);

	return 0;
};

static int __init kgr_patch_entry_code_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int __init kgr_patch_entry_code_init(void)
{
	int ret;

	ret =  kgr_patch_entry_code_kallsyms();
	if (ret)
		return ret;

	return kgr_patch_entry_code_relocate();
}

#endif /* IS_ENABLED(CONFIG_X86_64) && !IS_ENABLED(CONFIG_XEN) */
